
{*****************************************************************************}
{                                                                             }
{    Tnt LX Controls                                                          }
{      http://www.tntware.com/delphicontrols/lx/                              }
{        Version: 1.2.3                                                       }
{                                                                             }
{    Copyright (c) 2003-2006, Troy Wolbrink (troy.wolbrink@tntware.com)       }
{                                                                             }
{*****************************************************************************}

unit TntLXForms;

{$INCLUDE TntCompilers.inc}

interface

uses
  Forms, Classes, Controls, Windows, Messages, Menus, TntExtCtrls, ComCtrls,
  TntForms, TntComCtrls;

type
  PCustomForm = ^TCustomForm;

type
  TTntFormLX = class(TTntForm)
  private
    FOwnerForm: TCustomForm;
    FOwnerForm_IsParent: Boolean;
    FOwnerForm_SaveEnabled: Boolean;
    WindowList: Pointer;
    SaveActiveWindow: HWND;
    function GetBorderIcons: TBorderIcons;
    procedure SetBorderIcons(const Value: TBorderIcons);
    procedure SetOwnerForm(const Value: TCustomForm);
    procedure SetOwnerForm_IsParent(const Value: Boolean);
    function GetFormStyle: TFormStyle;
    function IsForm: Boolean;
    procedure SetFormStyle(const Value: TFormStyle);
    procedure CheckNotMainOrMDI;
  private
    FInMDISetMenu: Boolean;
    FClientInstance: TFarProc;
    FDefClientProc: TFarProc;
    procedure MyClientWndProc(var Msg: TMessage);
    procedure WMNCPaint(var Msg: TMessage); message WM_NCPAINT;
    function GetMenu: TMainMenu{TNT-ALLOW TMainMenu};
    procedure SetMenu(Value: TMainMenu{TNT-ALLOW TMainMenu});
    procedure CMMenuChanged(var Message: TMessage); message CM_MENUCHANGED;
    procedure WMInitMenuPopup(var Msg: TWMInitMenuPopup); message WM_INITMENUPOPUP;
  private
    FMDIControlBar: TTntControlBar;
    FMDIToolBar: TTntToolBar;
    FMDIStatusBar: TTntStatusBar;
    procedure SetMDIControlBar(const Value: TTntControlBar);
    procedure SetMDIStatusBar(const Value: TTntStatusBar);
    procedure SetMDIToolBar(const Value: TTntToolBar);
    procedure UpdateControlBar;
    procedure UpdateStatusBar;
    procedure UpdateToolBar;
  private
    FChildStatusBar: TTntStatusBar;
    FPrevStatusBarWinProc: Pointer;
    FSubClassedStatusBar: TTntStatusBar;
    FStatusBarWin32Proc_ObjectInstance: Pointer;
    FSaveStatusBarAlign: TAlign;
    FSaveStatusBarSizeGrip: Boolean;
    FSaveStatusBarRect: TRect;
    procedure StatusBarWin32Proc(var Message: TMessage);
  private
    FUpdateCount: Integer;
    FMDIChild_Activated: Boolean;
    procedure WMMDIActivate(var Message: TWMMDIActivate); message WM_MDIACTIVATE;
    procedure RedockToMDIForm(Force: Boolean);
    procedure MDIChild_UpdateMdiDockState(Activating: Boolean);
  protected
    AutoNilRef: PCustomForm;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateHandle; override;
    procedure CreateWnd; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure WndProc(var Message: TMessage); override;
    procedure Resizing(State: TWindowState); override;
  protected
    FTaskBarButton: Boolean;
    procedure UpdateApplicationMode(NewApplicationMode: Boolean);
    procedure WMSysCommand(var Message: TWMSysCommand); message WM_SYSCOMMAND;
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure WMActivateApp(var Message: TWMActivate); message WM_ACTIVATEAPP;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure SetTaskBarButton(const Value: Boolean);
    procedure DoClose(var Action: TCloseAction); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    procedure StartModal;
    procedure StopModal;
    property OwnerForm: TCustomForm read FOwnerForm write SetOwnerForm default nil;
    property OwnerForm_IsParent: Boolean read FOwnerForm_IsParent write SetOwnerForm_IsParent default False;
    procedure BeginUpdate;
    procedure EndUpdate;
    class procedure BeginClientUpdate;
    class procedure EndClientUpdate;
  published
    property BorderIcons: TBorderIcons read GetBorderIcons write SetBorderIcons stored IsForm
      default [biSystemMenu, biMinimize, biMaximize];
    property FormStyle: TFormStyle read GetFormStyle write SetFormStyle stored IsForm default fsNormal;
    property TaskBarButton: Boolean read FTaskBarButton write SetTaskBarButton default False;
    property Menu read GetMenu write SetMenu stored IsForm;
    property MDIControlBar: TTntControlBar read FMDIControlBar write SetMDIControlBar;
    property MDIToolBar: TTntToolBar read FMDIToolBar write SetMDIToolBar;
    property MDIStatusBar: TTntStatusBar read FMDIStatusBar write SetMDIStatusBar;
  end;

procedure CloseAllNonModalFormsExcept(ExceptForm: TTntFormLX; DoDestroy: Boolean = False);
procedure ConfirmCloseQueryOnAllNonModalForms;

var
  NeverShowApplicationWnd: Boolean;

implementation

uses
  SysUtils, RtlConsts, Consts, {$IFDEF COMPILER_7_UP} Themes, {$ENDIF} Graphics, CommCtrl,
  TntMenus, TntClasses, TntSysUtils, TntWindows;

resourcestring
  SCannotShowDisabledWindowAsModal = 'Cannot make a disabled window modal';

var
  NonModalFormList: TList;
  CloseMainFormWhenDone: Boolean;
  BringToFrontList: TList;
  WmBringToFront: Cardinal;

procedure CloseAllNonModalFormsExcept(ExceptForm: TTntFormLX; DoDestroy: Boolean = False);
var
  i: integer;
  Form: TTntFormLX;
begin
  for i := NonModalFormList.Count - 1 downto 0 do begin
    Form := TTntFormLX(NonModalFormList[i]);
    if Form <> ExceptForm then begin
      if DoDestroy then
        Form.Free
      else
        Form.Close;
    end;
  end;
end;

procedure ConfirmCloseQueryOnAllNonModalForms;
var
  i: integer;
  Form: TTntFormLX;
begin
  for i := NonModalFormList.Count - 1 downto 0 do begin
    Form := TTntFormLX(NonModalFormList[i]);
    if not Form.CloseQuery then
      Abort;
  end;
end;

function DlgIsModal(DlgHandle, OwnerWindow: Hwnd): Boolean;
var
  OwnedCorrectly: Boolean;
  Form: TWinControl;
begin
  OwnedCorrectly := IsWindow(DlgHandle)
                and (DlgHandle <> OwnerWindow) { not the same window }
                and (GetWindow(DlgHandle, GW_OWNER) = OwnerWindow) { correct owner }
                and (IsWindowEnabled(DlgHandle)); { window is enabled }
  if (not OwnedCorrectly) then
    Result := False
  else begin
    Form := FindControl(DlgHandle);
    if (Form is TCustomForm) and (fsModal in TCustomForm(Form).FormState) then
      Result := True { a Delphi modal TCustomForm }
    else
      { modal style }
      Result := ((GetWindowLong(DlgHandle, GWL_STYLE) and DS_MODALFRAME) = DS_MODALFRAME)
            and ((GetWindowLong(DlgHandle, GWL_EXSTYLE) and WS_EX_DLGMODALFRAME) = WS_EX_DLGMODALFRAME);
  end;
end;

function ModalHandle(OwnerHandle: Hwnd): Hwnd;
var
  Dlg: Hwnd;
begin
  // look for modal dialog
  Dlg := GetLastActivePopup(OwnerHandle);
  if DlgIsModal(Dlg, OwnerHandle) then
    Result := Dlg
  else
    Result := 0;
end;

function AppModalHandle: Hwnd;
begin
  Result := ModalHandle(Application.Handle);
end;

{ TTntFormLX }

constructor TTntFormLX.Create(AOwner: TComponent);
begin
  // standard construction technique (look at TForm.Create)
  GlobalNameSpace.BeginWrite;
  try
    CreateNew(AOwner);
    if (ClassType <> TTntFormLX) and not (csDesigning in ComponentState) then
    begin
      Include(FFormState, fsCreating);
      try
        if not InitInheritedComponent(Self, TTntFormLX) then
          raise EResNotFound.CreateFmt(SResNotFound, [ClassName]);
      finally
        Exclude(FFormState, fsCreating);
      end;
      if OldCreateOrder then DoCreate;
    end;
  finally
    GlobalNameSpace.EndWrite;
  end;
end;

destructor TTntFormLX.Destroy;
begin
  if FormStyle = fsMDIChild then
    BeginClientUpdate; { keeps other mdi children from flickering }
  try
    Destroying;
    TaskBarButton := False;
    if AutoNilRef <> nil then
      AutoNilRef^ := nil;
    inherited;
    {to make sure FClientInstance still out lives ClientHandle,
      it is freed after inherited destructor}
    Classes.FreeObjectInstance(FClientInstance);
    Classes.FreeObjectInstance(FStatusBarWin32Proc_ObjectInstance);
  finally
    if FormStyle = fsMDIChild then
      EndClientUpdate;
  end;
end;

procedure TTntFormLX.CreateParams(var Params: TCreateParams);
begin
  inherited;
  if (OwnerForm <> nil) then begin
    Params.WndParent := OwnerForm.Handle;
    // ws_child makes it position consistently
    // ws_popup makes it focus-able as a child
    if OwnerForm_IsParent then
      Params.Style := Params.Style or WS_CHILD or WS_POPUP;
  end;
  if TaskBarButton and (not (csDesigning in ComponentState)) then begin
    Params.ExStyle := Params.ExStyle or WS_EX_APPWINDOW and not WS_EX_TOOLWINDOW;
    Params.WndParent := 0;
  end;
end;

procedure TTntFormLX.CreateHandle;
begin
  inherited;
  if OwnerForm_IsParent and (OwnerForm <> nil) then
    Windows.SetParent(Handle, OwnerForm.Handle);    // windows.SetParent makes it contained
end;

procedure TTntFormLX.DoClose(var Action: TCloseAction);
begin
  if (FormStyle = fsMDIChild) then
    Action := caFree;
  inherited DoClose(Action);
  if (Application.MainForm = Self)
  and TaskBarButton
  and (not (csDesigning in ComponentState))
  and (Action <> caNone) then begin
    // check for deferred closing
    if (NonModalFormList.Count > 1) then begin
      // defer closing until all non-modal forms are closed
      Action := caNone;
      CloseMainFormWhenDone := True;
      Hide;
    end;
  end;
end;

procedure TTntFormLX.Loaded;
begin
  inherited;
  UpdateControlBar;
  UpdateToolBar;
  UpdateStatusBar;
  RedockToMDIForm(True);
end;

procedure TTntFormLX.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if (Operation = opRemove) then begin
    if (AComponent = FOwnerForm) then begin
      FOwnerForm := nil;
      if not (csDestroying in ComponentState) then
        RecreateWnd;
    end;
    if (AComponent = FMDIControlBar) then begin
      FMDIControlBar := nil;
      UpdateControlBar;
    end;
    if (AComponent = FMDIToolBar) then begin
      FMDIToolBar := nil;
      UpdateToolBar;
    end;
    if (AComponent = FMDIStatusBar) then begin
      FMDIStatusBar := nil;
      UpdateStatusBar;
    end;
  end;
  inherited;
end;

procedure TTntFormLX.WndProc(var Message: TMessage);
var
  i: integer;
begin
  if (Message.Msg = WmBringToFront)
  and (BringToFrontList.Count > 0) then begin
    for i := BringToFrontList.Count - 1 downto 0 do begin
      TTntFormLX(BringToFrontList[i]).BringToFront;
    end;
    BringToFrontList.Clear;
    SetActiveWindow(Message.WParam);
  end else begin
    {$IFDEF COMPILER_7_UP}
    if Message.Msg = WM_THEMECHANGED then
      UpdateControlBar;
    {$ENDIF}
    inherited
  end;
  if (Message.Msg = CM_SHOWINGCHANGED)
  and ((Application.MainForm = nil) or (Application.MainForm = Self)) then
    UpdateApplicationMode(TaskBarButton);
end;

procedure TTntFormLX.Resizing(State: TWindowState);
begin
  inherited;
  if (Application.MainForm is TTntFormLX)
  and (not TTntFormLX(Application.MainForm).FInMDISetMenu) then
    RedockToMDIForm(False);
end;

procedure TTntFormLX.WMActivateApp(var Message: TWMActivate);
begin
  inherited;
  if (Message.Active <> 0) and Visible and TaskBarButton and (not (csDesigning in ComponentState))
  and (not IsWindowEnabled(Handle)) and (AppModalHandle <> 0) then begin
    BringToFrontList.Add(Self);
    PostMessage(Handle, WmBringToFront, AppModalHandle, 0);
  end;
end;

var
  AssignedExitProc: Boolean;
  
procedure TntLXFormExitProc;
begin
  if (Application.MainForm <> nil)
  and (Application.MainForm.FormStyle = fsMDIForm) then
    Application.MainForm.Hide; { This way we don't get the flicker of seeing each client windows close. }
end;

procedure TTntFormLX.CMVisibleChanged(var Message: TMessage);
begin
  { CM_VISIBLECHANGED is one of the first messages received after Application.MainForm is assigned. }
  if (Application.MainForm = Self) then
    UpdateApplicationMode(TaskBarButton);
  if (Self = Application.MainForm) and Visible and (not AssignedExitProc) then begin
    AddExitProc(TntLXFormExitProc); // this should come after AddExitProc(DoneApplication) in TApplication.Run;
    AssignedExitProc := True;
  end;
  inherited;
end;

procedure TTntFormLX.WMSysCommand(var Message: TWMSysCommand);
begin
  with Message do begin
    if (CmdType and $FFF0 = SC_MINIMIZE) and (Application.MainForm = Self)
    and TaskBarButton and (not (csDesigning in ComponentState)) then
      DefaultHandler(Message) { bypass TApplication stuff for minimizing with hidden app window }
    else
      inherited;
  end;
end;

procedure TTntFormLX.WMActivate(var Message: TWMActivate);
begin
  inherited;
  if (Message.Active = WA_ACTIVE)
  and TaskBarButton and (not (csDesigning in ComponentState)) then
  begin
    if (not IsWindowEnabled(Handle)) and (AppModalHandle <> 0) then begin
      {when clicking on taskbar, this keeps any modal forms from being hidden}
      SetActiveWindow(AppModalHandle);
    end;
    UpdateApplicationMode(True);
  end;
end;

type
  TFlashWindowEx = function(var pfwi: FLASHWINFO): BOOL; stdcall;

var
  FlashWindowEx: TFlashWindowEx;

procedure LoadFlashWindowEx;
var
  User32: THandle;
begin
  User32 := WinCheckH(Tnt_LoadLibraryW('user32.dll'));
  try
    // FlashWindowEx is not available on Windows 95/NT4
    FlashWindowEx := GetProcAddress(User32, 'FlashWindowEx');
  finally
    FreeLibrary(User32);
  end;
end;

procedure TTntFormLX.WMSetCursor(var Message: TWMSetCursor);

  procedure FlashDialog(Handle: HWND);
  var
    FlashInfo: TFLASHWINFO;
  begin
    if not Assigned(FlashWindowEx) then
      FlashWindow(Handle, True)
    else begin
      FlashInfo.cbSize := 4 * SizeOf(Cardinal) + SizeOf(HWND);
      FlashInfo.hwnd := Handle;
      FlashInfo.dwFlags := FLASHW_CAPTION;
      FlashInfo.uCount := 3;
      FlashInfo.dwTimeout := 70;
      FlashWindowEx(FlashInfo);
    end;
  end;

var
  AppActive: Boolean;

  procedure SendFocusToModalDialog(Handle: THandle; DoBeep: Boolean);
  begin
    SetActiveWindow(Handle);
    if AppActive then begin
      if DoBeep then
        Beep;  {audibly alert user of focus shift}
      FlashDialog(Handle);
    end;
  end;

begin
  AppActive := Application.Active;
  inherited;
  with Message do begin
    if  (CursorWnd = Handle) and (MouseMsg = WM_LBUTTONDOWN) and (not IsWindowEnabled(Handle)) then
    begin
      {clicking on this disabled window... check for a modal window.}
      if ModalHandle(Handle) <> 0 then
        SendFocusToModalDialog(ModalHandle(Handle), True)
      else if AppModalHandle <> 0 then
        SendFocusToModalDialog(AppModalHandle, False);
    end;
  end;
end;

function TTntFormLX.IsForm: Boolean;
begin
  Result := not IsControl;
end;

function TTntFormLX.GetBorderIcons: TBorderIcons;
begin
  Result := inherited BorderIcons;
end;

procedure TTntFormLX.SetBorderIcons(const Value: TBorderIcons);
begin
  if Value <> BorderIcons then begin
    inherited BorderIcons := Value;
    {$IFDEF COMPILER_10_UP}
    if (FormStyle = fsMDIChild) and (Value = []) then
      RecreateWnd; // Delphi 2006 will try to update buttons w/o recreate and this doesn't work
    {$ENDIF}
  end;
end;

function TTntFormLX.GetFormStyle: TFormStyle;
begin
  Result := inherited FormStyle;
end;

procedure TTntFormLX.SetFormStyle(const Value: TFormStyle);
begin
  if Value <> FormStyle then begin
    if ((Value in [fsMDIForm, fsMDIChild]) or (Application.MainForm = Self))
    and ((OwnerForm <> nil) or (OwnerForm_IsParent = True)) then
      raise ETntInternalError.Create('Internal Error: An owned form can not be a main or MDI form.');
    if (Value = fsMDIChild) and TaskBarButton then
      raise ETntInternalError.Create('Internal Error: An MDI child form can not have a taskbar button.');
    inherited FormStyle := Value;
    RedockToMDIForm(True);
  end;
end;

// --- OwnerForm / StartModal / StopModal ---

procedure TTntFormLX.CheckNotMainOrMDI;
begin
  if ((Application.MainForm = Self) or (FormStyle in [fsMDIForm, fsMDIChild])) then
    raise ETntInternalError.Create('Internal Error: This operation is not valid for a main or MDI form.');
  if TaskBarButton then
    raise ETntInternalError.Create('Internal Error: This operation is not valid for a form with a TaskBar button.');
end;

procedure TTntFormLX.SetOwnerForm(const Value: TCustomForm);
begin
  if FOwnerForm <> Value then begin
    CheckNotMainOrMDI;
    FOwnerForm := Value;
    if FOwnerForm <> nil then
      FOwnerForm.FreeNotification(Self);
    RecreateWnd;
  end;
end;

procedure TTntFormLX.SetOwnerForm_IsParent(const Value: Boolean);
begin
  if FOwnerForm_IsParent <> Value then begin
    CheckNotMainOrMDI;
    FOwnerForm_IsParent := Value;
    if OwnerForm <> nil then
      RecreateWnd;
  end;
end;

procedure TTntFormLX.StartModal;
begin
  CheckNotMainOrMDI;
  if Visible then
    raise EInvalidOperation.Create(SCannotShowModal);
  if (not Enabled) then
    raise ETntGeneralError.Create(SCannotShowDisabledWindowAsModal);
  if OwnerForm <> nil then begin
    FOwnerForm_SaveEnabled := OwnerForm.Enabled;
    OwnerForm.Enabled := False;
  end else begin
    SaveActiveWindow := GetActiveWindow;
    WindowList := DisableTaskWindows(0);
  end;
  FFormState := FFormState + [fsModal];
  Show;
end;

procedure TTntFormLX.StopModal;
begin
  CheckNotMainOrMDI;
  if GetActiveWindow <> Handle then SaveActiveWindow := 0;
  Hide;
  FFormState := FFormState - [fsModal];

  if OwnerForm <> nil then begin
    OwnerForm.Enabled := FOwnerForm_SaveEnabled;
    if OwnerForm.CanFocus then
      OwnerForm.SetFocus;
  end else begin
    if Assigned(WindowList) then begin
      EnableTaskWindows(WindowList);
      WindowList := nil;
    end;
    if SaveActiveWindow <> 0 then begin
      SetActiveWindow(SaveActiveWindow);
      SaveActiveWindow := 0;
    end;
  end;
end;

// --- TaskBarButton ---

procedure TTntFormLX.UpdateApplicationMode(NewApplicationMode: Boolean);
begin
  if (NewApplicationMode or NeverShowApplicationWnd)
  and (not (csDesigning in ComponentState)) then
    ShowWindow(Application.Handle, SW_HIDE) { hide application window / task bar button }
  else
    ShowWindow(Application.Handle, SW_SHOW) { show application window / task bar button }
end;

procedure TTntFormLX.SetTaskBarButton(const Value: Boolean);

  procedure AddForm(Form: TTntFormLX);
  begin
    NonModalFormList.Add(Form);
  end;

  procedure RemoveForm(Form: TTntFormLX);
  begin
    NonModalFormList.Remove(Form);
    if CloseMainFormWhenDone
    and (NonModalFormList.Count = 1) { the main form is all that's left }
    and (Application.MainForm = NonModalFormList[0]) then
      Application.MainForm.Close;
  end;

begin
  if FTaskBarButton <> Value then begin
    if FTaskBarButton then
      RemoveForm(Self);
    if (FormStyle = fsMDIChild) or (OwnerForm <> nil) or (OwnerForm_IsParent = True) then
      raise ETntGeneralError.Create('A MDI child or a contained form can not have its own task bar button.');
    FTaskBarButton := Value;
    if FTaskBarButton then
      AddForm(Self);
    if  (not (csDestroying in ComponentState))
    and (not (csDesigning in ComponentState)) then begin
      RecreateWnd;
      if Application.MainForm = Self then
        UpdateApplicationMode(TaskBarButton);
    end;
  end;
end;

// --- BeginUpdate / EndUpdate

procedure TTntFormLX.BeginUpdate;
begin
  if HandleAllocated and (FUpdateCount = 0) then
    SendMessage(Handle, WM_SETREDRAW, 0, 0);
  Inc(FUpdateCount);
end;

procedure TTntFormLX.EndUpdate;
begin
  if FUpdateCount > 0 then
    Dec(FUpdateCount);
  if HandleAllocated and (FUpdateCount = 0) then begin
    SendMessage(Handle, WM_SETREDRAW, 1, 0);
    RedrawWindow(Handle, nil, 0,
      RDW_FRAME or RDW_INVALIDATE or RDW_ALLCHILDREN or RDW_NOINTERNALPAINT);
  end;
end;

function MainClientHandle: THandle;
begin
  if (Application.MainForm <> nil) then
    Result := Application.MainForm.ClientHandle
  else
    Result := 0;
end;

var
  ClientUpdateCount: Integer;
  ClientUpdateLocked: Boolean;

class procedure TTntFormLX.BeginClientUpdate;
begin
  if (MainClientHandle <> 0) and (ClientUpdateCount = 0) then begin
    ClientUpdateLocked := LockWindowUpdate(Application.MainForm.Handle);
    SendMessage(MainClientHandle, WM_SETREDRAW, 0, 0);
  end;
  Inc(ClientUpdateCount);
end;

class procedure TTntFormLX.EndClientUpdate;
begin
  if ClientUpdateCount > 0 then
    Dec(ClientUpdateCount);
  if (ClientUpdateCount = 0) then begin
    if (MainClientHandle <> 0) then begin
      SendMessage(MainClientHandle, WM_SETREDRAW, 1, 0);
      RedrawWindow(MainClientHandle, nil, 0,
        RDW_FRAME or RDW_INVALIDATE or RDW_ALLCHILDREN or RDW_NOINTERNALPAINT);
      Application.MainForm.Refresh;
    end;
    if ClientUpdateLocked then
      LockWindowUpdate(0);
  end;
end;

// --- No Owner Drawn Top Level Menu Items ---

procedure TTntFormLX.CMMenuChanged(var Message: TMessage);
begin
  NoOwnerDrawTopLevelItems(Menu);
  inherited;
end;

function TTntFormLX.GetMenu: TMainMenu{TNT-ALLOW TMainMenu};
begin
  Result := inherited Menu;
end;

procedure TTntFormLX.SetMenu(Value: TMainMenu{TNT-ALLOW TMainMenu});
{This is needed to catch initial assignment to Menu property when
Form is read from Stream. Without this, there is a slight
flicker the first time the mouse hovers over the menu}
begin
  NoOwnerDrawTopLevelItems(Value);
  inherited Menu := Value;
end;

procedure TTntFormLX.WMInitMenuPopup(var Msg: TWMInitMenuPopup);
{If the window menu is about to drop down, tell Windows to refresh it. This
adds any open MDI children to menu}
begin
  if (FormStyle = fsMDIForm) and Assigned(WindowMenu) and (WindowMenu.Handle = Msg.MenuPopup) then
    SendMessage(ClientHandle, WM_MDIREFRESHMENU, 0, 0);
  inherited;
end;

//-----------------------

{Everything from this point on deals with trying to get the mainmenu right when
dealing with maximized MDI children. The main idea is to bracket calls to
WM_MDISETMENU with a restore/maximize of the active MDI child. This ensures
that the icon and window buttons are correctly added to the menu bar. Since the
restore/maximize results in ugly flicker in the menu bar, there is also code
here to suppress repainting of the non-client area until the re-maximize is
complete}

procedure TTntFormLX.CreateWnd;
begin
  inherited;
  if (FormStyle = fsMDIForm) then begin
    {FClientInstance created here (rather than in a TMdiMainForm.Create) because the
    form's window handle might be created during processing of TForm.Create}
    if FClientInstance = nil then
      FClientInstance:= Classes.MakeObjectInstance(MyClientWndProc);
    {Need to catch WM_MDISETMENU messages sent to ClientHandle, so subclass it here}
    FDefClientProc:= TFarProc(SetWindowLong(ClientHandle, GWL_WNDPROC,
      integer(FClientInstance)));
  end;
end;

procedure TTntFormLX.MyClientWndProc(var Msg: TMessage);

  procedure InvalidateNonClientArea;
  var
    R: TRect;
  begin
    R := Rect(0, -1, Width, 0);
    RedrawWindow(Handle, @R, 0, RDW_INVALIDATE or RDW_FRAME);
  end;

var
  ActiveChild: TForm{TNT-ALLOW TForm};
  Size: integer;
begin
  Size:= 0; // suppress warnings
  ActiveChild:= nil;
  if Msg.Msg = WM_MDISETMENU then begin
    ActiveChild := ActiveMDIChild;
    if Assigned(ActiveChild) and (ActiveChild.WindowState = wsMaximized) then begin
      FInMDISetMenu:= true;
      {The following (sending WM_SIZE messages directly to the ActiveChild) is
      cribbed from the VCL's MergeMenu routine prior to D5.
      It seems to work pretty well to convince Windows that the ActiveChild is
      not maximized, while minimizing any necessary repainting of the
      ActiveChild}
      with ActiveChild do
        Size := (ClientWidth) + (ClientHeight shl 16);
      SendMessage(ActiveChild.Handle, WM_SIZE, SIZE_RESTORED, Size);
    end else
      ActiveChild:= nil;
  end;
  try
    with Msg do
      Result:= CallWindowProc(FDefClientProc, ClientHandle, Msg, wParam, lParam);
  finally
    if Assigned(ActiveChild) then begin
      {re-maximize the ActiveChild}
      SendMessage(ActiveChild.Handle, WM_SIZE, SIZE_MAXIMIZED, Size);
      FInMDISetMenu:= false;
      InvalidateNonClientArea; { take care of the WM_NCPAINT messages we just ignored }
    end;
  end;
end;

procedure TTntFormLX.WMNCPaint(var Msg: TMessage);
begin
  if (FormStyle <> fsMDIForm) or (not FInMDISetMenu) then
    inherited
  else begin
    {We are inside a WM_MDISETMENU, so ignore the message}
  end;
end;

//---------------- MDI Control Bar, MDI Tool Bar, MDI Status Bar ------------------------

procedure TTntFormLX.UpdateControlBar;
begin
  {$IFDEF COMPILER_7_UP}
  if (MDIControlBar <> nil) and (not (csDesigning in ComponentState)) then begin
    if ThemeServices.ThemesEnabled then
      MDIControlBar.Color := clMenuBar
    else
      MDIControlBar.Color := clMenu;
  end;
  {$ENDIF}
end;

procedure TTntFormLX.UpdateToolBar;
begin
  //
end;

procedure TTntFormLX.UpdateStatusBar;
begin
  if FSubClassedStatusBar <> MDIStatusBar then begin
    // unsubclass old one
    if FSubClassedStatusBar <> nil then begin
      if FSubClassedStatusBar.HandleAllocated then
        SetWindowLong(FSubClassedStatusBar.Handle, GWL_WNDPROC, Integer(FPrevStatusBarWinProc));
      FPrevStatusBarWinProc := nil;
      FSubClassedStatusBar := nil;
    end;
    if MDIStatusBar <> nil then begin
      if FStatusBarWin32Proc_ObjectInstance = nil then
        FStatusBarWin32Proc_ObjectInstance := MakeObjectInstance(StatusBarWin32Proc);
      FPrevStatusBarWinProc := Pointer(SetWindowLong(MDIStatusBar.Handle, GWL_WNDPROC, Integer(FStatusBarWin32Proc_ObjectInstance)));
      FSubClassedStatusBar := MDIStatusBar;
    end;
  end;
end;

procedure TTntFormLX.RedockToMDIForm(Force: Boolean);
begin
  if  (not (csLoading in ComponentState))
  and (FormStyle = fsMDIChild)
  and (Application.MainForm <> nil)
  and (Application.MainForm.ActiveMDIChild = Self) then begin
    if Force then BeginClientUpdate;
    try
      if Force then begin
        MDIChild_UpdateMdiDockState(WindowState <> wsMaximized);
      end;
      MDIChild_UpdateMdiDockState(WindowState = wsMaximized);
    finally
      if Force then EndClientUpdate;
    end;
  end;
end;

procedure TTntFormLX.SetMDIControlBar(const Value: TTntControlBar);
begin
  if (FMDIControlBar <> Value) then begin
    FMDIControlBar := Value;
    if MDIControlBar <> nil then
      MDIControlBar.FreeNotification(Self);
    UpdateControlBar;
    RedockToMDIForm(True);
  end;
end;

procedure TTntFormLX.SetMDIToolBar(const Value: TTntToolBar);
begin
  if (FMDIToolBar <> Value) then begin
    FMDIToolBar := Value;
    if MDIToolBar <> nil then
      MDIToolBar.FreeNotification(Self);
    UpdateToolBar;
    RedockToMDIForm(True);
  end;
end;

procedure TTntFormLX.SetMDIStatusBar(const Value: TTntStatusBar);
begin
  if (FMDIStatusBar <> Value) then begin
    FMDIStatusBar := Value;
    if MDIStatusBar <> nil then
      MDIStatusBar.FreeNotification(Self);
    UpdateStatusBar;
    RedockToMDIForm(True);
  end;
end;

//---------- MDI Child / Update MDI Dock State (Control Bar, Tool Bar, Status Bar ) -----------

procedure TTntFormLX.WMMDIActivate(var Message: TWMMDIActivate);
var
  BecomingActive: Boolean;
  LosingActive: Boolean;
begin
  inherited;
  if (FormStyle = fsMDIChild) then
  begin
    BecomingActive := Message.ActiveWnd = Handle;
    LosingActive := Message.DeactiveWnd = Handle;
    if (BecomingActive) then begin
      UpdateWindowState;
      MDIChild_UpdateMdiDockState(WindowState = wsMaximized)
    end else if (LosingActive) then
      MDIChild_UpdateMdiDockState(False);
  end;
end;

function GetFirstPanelWidth(SB: TTntStatusBar): Integer;
begin
  if (SB.SimplePanel) or (SB.Panels.Count = 0) then
    Result := 100
  else
    Result := SB.Panels[0].Width;
end;

function GetTotalPanelWidth(SB: TTntStatusBar): Integer;
var
  i: integer;
begin
  if (SB.SimplePanel) or (SB.Panels.Count = 0) then
    Result := GetFirstPanelWidth(SB)
  else begin
    Result := 0;
    for i := 0 to SB.Panels.Count - 1 do
      Inc(Result, SB.Panels[i].Width);
  end;
end;

type
  TAccessCollection = class(TCollection);

procedure TTntFormLX.MDIChild_UpdateMdiDockState(Activating: Boolean);
var
  MainControlBar: TTntControlBar;
  MainToolBar: TTntToolBar;
  MainStatusBar: TTntStatusBar;
begin
  if csLoading in ComponentState then
    exit;

  if (FormStyle <> fsMDIChild) then
    raise ETntInternalError.Create('Internal Error: MDIChild_UpdateMdiDockState can only be called from MDI child forms.');

  if (Activating <> FMDIChild_Activated)
  and (Application.MainForm is TTntFormLX) then begin
    BeginClientUpdate;
    try
      with TTntFormLX(Application.MainForm) do begin
        MainControlBar := MDIControlBar;
        MainToolBar := MDIToolBar;
        MainStatusBar := MDIStatusBar;
      end;

      // dock tool bar
      if  (MDIToolBar    <> nil) and (MainToolBar    <> nil)
      and (MDIControlBar <> nil) and (MainControlBar <> nil) then begin
        MDIToolBar.Hide;
        if Activating then begin
          MDIToolBar.Parent := MainControlBar;
          MDIToolBar.Left := MainToolBar.Left + 1;
        end else begin
          MDIToolBar.Parent := MDIControlBar;
          MDIToolBar.Left := 0;
        end;
        MDIControlBar.Top := 0;
        MDIToolBar.Show;
      end;

      // dock status bar
      if (MDIStatusBar <> nil)
      and (MainStatusBar <> nil) then begin
        if Activating then begin
          with MDIStatusBar do begin
            FSaveStatusBarRect := Rect(Left, Top, Left + Width, Top + Height);
            FSaveStatusBarAlign := Align;
            FSaveStatusBarSizeGrip := SizeGrip;
            SizeGrip := False;
            Align := alNone;
            Top := 0;
            Height := MainStatusBar.Height;
            Left := GetFirstPanelWidth(MainStatusBar) + 2;
            Width := GetTotalPanelWidth(MDIStatusBar) - 2;
            Parent := MainStatusBar;
          end;
          TTntFormLX(Application.MainForm).FChildStatusBar := MDIStatusBar;
          TAccessCollection(MainStatusBar.Panels).Update(nil);
        end else begin
          { de-activating }
          with MDIStatusBar do begin
            Parent := Self;
            Align := FSaveStatusBarAlign;
            with FSaveStatusBarRect do
              SetBounds(Left, Top, Right - Left, Bottom - Top);
            SizeGrip := FSaveStatusBarSizeGrip;
          end;
          TTntFormLX(Application.MainForm).FChildStatusBar := nil;
          TAccessCollection(MainStatusBar.Panels).Update(nil);
        end;
      end;
      FMDIChild_Activated := Activating;
    finally
      EndClientUpdate;
    end;
  end;
end;

//----------------------------------------------

procedure TTntFormLX.StatusBarWin32Proc(var Message: TMessage);
var
  ClientBarWidth: Integer;
  FirstPanelWidth: Integer;
  P: PInteger;
  i: integer;
begin
  if FChildStatusBar <> nil then
    ClientBarWidth := GetTotalPanelWidth(FChildStatusBar)
  else begin
    ClientBarWidth := 0;
  end;
  if (FChildStatusBar <> nil) and (Message.Msg = SB_SETPARTS) then begin
    P := PInteger(Message.lParam);
    for i := 1 to Message.WParam do begin
      if P^ <> -1 then
        P^ := P^ + ClientBarWidth;
      Inc(P);
    end;
  end;
  with Message do begin
    Result := CallWindowProc(FPrevStatusBarWinProc, FSubClassedStatusBar.Handle, Msg, wParam, lParam);
  end;
  if (FChildStatusBar <> nil) and (Message.Msg = WM_PAINT) then begin
    FirstPanelWidth := GetFirstPanelWidth(MDIStatusBar);
    with MDIStatusBar.Canvas do begin
      Pen.Color := MDIStatusBar.Color;
      MoveTo(FirstPanelWidth + -1, 0);
      LineTo(FirstPanelWidth + -1, MDIStatusBar.ClientHeight);
      MoveTo(FirstPanelWidth + 0, 0);
      LineTo(FirstPanelWidth + 0, MDIStatusBar.ClientHeight);
      Pen.Color := FChildStatusBar.Color;
      MoveTo(FirstPanelWidth + 1, 0);
      LineTo(FirstPanelWidth + 1, MDIStatusBar.ClientHeight);
      MoveTo(FirstPanelWidth + ClientBarWidth, 0);
      LineTo(FirstPanelWidth + ClientBarWidth, MDIStatusBar.ClientHeight);
      Pen.Color := MDIStatusBar.Color;
      MoveTo(FirstPanelWidth + ClientBarWidth + 1, 0);
      LineTo(FirstPanelWidth + ClientBarWidth + 1, MDIStatusBar.ClientHeight);
      MoveTo(FirstPanelWidth + ClientBarWidth + 2, 0);
      LineTo(FirstPanelWidth + ClientBarWidth + 2, MDIStatusBar.ClientHeight - 1);
      Pen.Color := clWhite;
      MoveTo(FirstPanelWidth - 1, 2);
      LineTo(FirstPanelWidth - 1, MDIStatusBar.ClientHeight);
      Pen.Color := clDkGray;
      MoveTo(FirstPanelWidth + ClientBarWidth + 2, 2);
      LineTo(FirstPanelWidth + ClientBarWidth + 2, MDIStatusBar.ClientHeight - 1);
    end;
  end;
end;

//----------------------------------------------

initialization
  LoadFlashWindowEx;
  NonModalFormList := TList.Create;
  CloseMainFormWhenDone := False;
  BringToFrontList := TList.Create;
  WmBringToFront := RegisterWindowMessage('Tnt.BringToFront');

finalization
  NonModalFormList.Free;
  BringToFrontList.Free;

end.
