
{*****************************************************************************}
{                                                                             }
{    Tnt LX Controls                                                          }
{      http://www.tntware.com/delphicontrols/lx/                              }
{        Version: 1.2.3                                                       }
{                                                                             }
{    Copyright (c) 2003-2006, Troy Wolbrink (troy.wolbrink@tntware.com)       }
{                                                                             }
{*****************************************************************************}

unit TntLXRichEdits;

{$INCLUDE TntCompilers.inc}

interface

uses
  Classes, Windows, Messages, Controls, TntComCtrls, TntLXUtils, TntDBCtrls;

type
  TRichEditLXVarBlock = record
    AutoUrlDetect: Boolean;
    OnLinkClick: TStringEvent;
    LastLinkTime: Cardinal;
    SaveHint: WideString;
    ShowLinkHint: Boolean;
  end;

  TTntCustomRichEditLX = class(TTntCustomRichEdit)
  private
    FVarBlock: TRichEditLXVarBlock;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure SetAutoUrlDetect(const Value: Boolean);
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure SetShowLinkHint(const Value: Boolean);
  protected
    procedure CreateWindowHandle(const Params: TCreateParams); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DoLinkClick(const LinkText: WideString); dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateWnd; override;
    //--
    property AutoUrlDetect: Boolean read FVarBlock.AutoUrlDetect write SetAutoUrlDetect default True;
    property ShowLinkHint: Boolean read FVarBlock.ShowLinkHint write SetShowLinkHint default True;
    property OnLinkClick: TStringEvent read FVarBlock.OnLinkClick write FVarBlock.OnLinkClick;
  end;

  TTntRichEditLX = class(TTntCustomRichEditLX)
  published
    property AutoUrlDetect;
    property ShowLinkHint;
    property OnLinkClick;
  published
    property Align;
    property Alignment;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind default bkNone;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property HideScrollBars;
    property ImeMode;
    property ImeName;
    property Constraints;
    property Lines;
    property MaxLength;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PlainText;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property WantTabs;
    property WantReturns;
    property WordWrap;
    property OnChange;
    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnProtectChange;
    property OnResizeRequest;
    property OnSaveClipboard;
    property OnSelectionChange;
    property OnStartDock;
    property OnStartDrag;
  end;

  TTntDBRichEditLX = class(TTntDBRichEdit)
  private
    FVarBlock: TRichEditLXVarBlock;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure SetAutoUrlDetect(const Value: Boolean);
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure SetShowLinkHint(const Value: Boolean);
  protected
    procedure CreateWindowHandle(const Params: TCreateParams); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DoLinkClick(const LinkText: WideString); dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateWnd; override;
  published
    property AutoUrlDetect: Boolean read FVarBlock.AutoUrlDetect write SetAutoUrlDetect default True;
    property ShowLinkHint: Boolean read FVarBlock.ShowLinkHint write SetShowLinkHint default True;
    property OnLinkClick: TStringEvent read FVarBlock.OnLinkClick write FVarBlock.OnLinkClick;
  end;

implementation

uses
  Forms, RichEdit, TntWindows, TntSysUtils, TntControls, TntForms, TntLXSafeWinInet;

procedure TntRichEditLX_AfterInherited_Create(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock);
begin
  with VarBlock do begin
    AutoUrlDetect := True;
    ShowLinkHint := True;
  end;
end;

var
  FRichEdit40Module: THandle = 0;

function IsRichEdit40Available: Boolean;
const
  RICHED40_DLL = 'MSFTEDIT.DLL';
begin
  if FRichEdit40Module = 0 then
    FRichEdit40Module := Tnt_LoadLibraryW(RICHED40_DLL);
  Result := FRichEdit40Module <> 0;
end;

const
  MSFTEDIT_CLASSW = 	'RICHEDIT50W';

function TntRichEditLX_CreateWindowHandle(Control: TTntCustomRichEdit; const Params: TCreateParams): Boolean;
begin
  Result := Win32PlatformIsUnicode and IsRichEdit40Available;
  if Result then
    CreateUnicodeHandle(Control, Params, MSFTEDIT_CLASSW);
end;

procedure TntRichEditLX_AfterInherited_CreateWnd(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock);
var
  EventMask: Longint;
begin
  with Control, VarBlock do begin
    if AutoUrlDetect then begin
      SendMessage(Handle, EM_AUTOURLDETECT, 1, 0);
      EventMask := SendMessage(Handle, EM_GETEVENTMASK, 0, 0) or ENM_Link;
      SendMessage(Handle, EM_SETEVENTMASK, 0, EventMask);
    end;
  end;
end;

type TAccessWinControl = class(TWinControl);

procedure TntRichEditLX_SetAutoUrlDetect(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock; const Value: Boolean);
begin
  with Control, VarBlock do begin
    if AutoUrlDetect <> Value then begin
      AutoUrlDetect := Value;
      TAccessWinControl(Control).RecreateWnd;
    end;
  end;
end;

procedure TntRichEditLX_DoLinkClick(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock; const LinkText: WideString);
begin
  with Control, VarBlock do begin
    if Assigned(OnLinkClick) then
      OnLinkClick(LinkText)
    else
      OpenUserUrl(LinkText);
  end;
end;

procedure TntRichEditLX_SetLinkHint(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock; const Value: WideString);
begin
  with Control, VarBlock do begin
    if SaveHint = '' then
      SaveHint := Hint; { remember original hint }
    if Value = '' then begin
      Hint := SaveHint; { restore original hint }
      SaveHint := '';
    end else if ShowLinkHint then begin
      Hint := Value;
      if TntApplication.Hint <> Hint then
        Application.CancelHint;
    end;
  end;
end;

procedure TntRichEditLX_SetShowLinkHint(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock; const Value: Boolean);
begin
  with Control, VarBlock do begin
    ShowLinkHint := Value;
    if not ShowLinkHint then
      TntRichEditLX_SetLinkHint(Control, VarBlock, '');
  end;
end;

type
  TAccessTntCustomRichEdit = class(TTntCustomRichEdit);

procedure TntRichEditLX_AfterInherited_CNNotify(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock; var Message: TWMNotify);
type
  PENLink = ^TENLink;

  procedure CheckLinkMessage(EmLink: PENLink);
  var
    LinkText: WideString;
    LinkStart, LinkLength: Integer;
  begin
    with TAccessTntCustomRichEdit(Control), VarBlock do begin
      LinkStart := EmulatedCharPos(EMLink.chrg.cpMin) + 1;
      LinkLength := EmulatedCharPos(EMLink.chrg.cpMax) - EmulatedCharPos(EMLink.chrg.cpMin);
      LinkText := Copy(Text, LinkStart, LinkLength);
      case EMLink.msg of
        WM_LBUTTONUP:
          TntRichEditLX_DoLinkClick(Control, VarBlock, LinkText);
        WM_MOUSEMOVE:
          begin
            TntRichEditLX_SetLinkHint(Control, VarBlock, LinkText);
            LastLinkTime := GetTickCount;
          end;
      end;
    end;
  end;

begin
  if Message.NMHdr^.code = EN_LINK then
    CheckLinkMessage(PENLink(Message.NMHdr));
end;

procedure TntRichEditLX_AfterInherited_MouseMove(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock);
begin
  with VarBlock do begin
    if (GetTickCount - LastLinkTime) > 100 then
      TntRichEditLX_SetLinkHint(Control, VarBlock, '');
  end;
end;

procedure TntRichEditLX_AfterInherited_CMMouseLeave(Control: TTntCustomRichEdit; var VarBlock: TRichEditLXVarBlock);
begin
  with VarBlock do begin
    if (GetTickCount - LastLinkTime) > 100 then
      TntRichEditLX_SetLinkHint(Control, VarBlock, '');
  end;
end;

{ TTntCustomRichEditLX }

constructor TTntCustomRichEditLX.Create(AOwner: TComponent);
begin
  inherited;
  TntRichEditLX_AfterInherited_Create(Self, FVarBlock);
end;

procedure TTntCustomRichEditLX.CreateWindowHandle(const Params: TCreateParams);
begin
  if not TntRichEditLX_CreateWindowHandle(Self, Params) then
    inherited;
end;

procedure TTntCustomRichEditLX.CreateWnd;
begin
  inherited;
  TntRichEditLX_AfterInherited_CreateWnd(Self, FVarBlock);
end;

procedure TTntCustomRichEditLX.SetAutoUrlDetect(const Value: Boolean);
begin
  TntRichEditLX_SetAutoUrlDetect(Self, FVarBlock, Value);
end;

procedure TTntCustomRichEditLX.DoLinkClick(const LinkText: WideString);
begin
  TntRichEditLX_DoLinkClick(Self, FVarBlock, LinkText);
end;

procedure TTntCustomRichEditLX.SetShowLinkHint(const Value: Boolean);
begin
  TntRichEditLX_SetShowLinkHint(Self, FVarBlock, Value);
end;

procedure TTntCustomRichEditLX.CNNotify(var Message: TWMNotify);
begin
  inherited;
  TntRichEditLX_AfterInherited_CNNotify(Self, FVarBlock, Message);
end;

procedure TTntCustomRichEditLX.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  TntRichEditLX_AfterInherited_MouseMove(Self, FVarBlock);
end;

procedure TTntCustomRichEditLX.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  TntRichEditLX_AfterInherited_MouseMove(Self, FVarBlock);
end;

{ TTntDBRichEditLX }

constructor TTntDBRichEditLX.Create(AOwner: TComponent);
begin
  inherited;
  TntRichEditLX_AfterInherited_Create(Self, FVarBlock);
end;

procedure TTntDBRichEditLX.CreateWindowHandle(const Params: TCreateParams);
begin
  if not TntRichEditLX_CreateWindowHandle(Self, Params) then
    inherited;
end;

procedure TTntDBRichEditLX.CreateWnd;
begin
  inherited;
  TntRichEditLX_AfterInherited_CreateWnd(Self, FVarBlock);
end;

procedure TTntDBRichEditLX.SetAutoUrlDetect(const Value: Boolean);
begin
  TntRichEditLX_SetAutoUrlDetect(Self, FVarBlock, Value);
end;

procedure TTntDBRichEditLX.DoLinkClick(const LinkText: WideString);
begin
  TntRichEditLX_DoLinkClick(Self, FVarBlock, LinkText);
end;

procedure TTntDBRichEditLX.SetShowLinkHint(const Value: Boolean);
begin
  TntRichEditLX_SetShowLinkHint(Self, FVarBlock, Value);
end;

procedure TTntDBRichEditLX.CNNotify(var Message: TWMNotify);
begin
  inherited;
  TntRichEditLX_AfterInherited_CNNotify(Self, FVarBlock, Message);
end;

procedure TTntDBRichEditLX.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  TntRichEditLX_AfterInherited_MouseMove(Self, FVarBlock);
end;

procedure TTntDBRichEditLX.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  TntRichEditLX_AfterInherited_MouseMove(Self, FVarBlock);
end;

initialization

finalization
  if FRichEdit40Module <> 0 then
    FreeLibrary(FRichEdit40Module);

end.
