
{*****************************************************************************}
{                                                                             }
{    Tnt LX Controls                                                          }
{      http://www.tntware.com/delphicontrols/lx/                              }
{        Version: 1.2.3                                                       }
{                                                                             }
{    Copyright (c) 2003-2006, Troy Wolbrink (troy.wolbrink@tntware.com)       }
{                                                                             }
{*****************************************************************************}

unit TntLXSafeWinInet;

{$INCLUDE TntCompilers.inc}

// This unit replaces the WinInet unit.
//   It safely checks to make sure IE 4.0 is installed before trying to load WinInet.dll

{TNT-WARN WinInet}

interface

uses
  Windows;

function InternetIsConnected: Boolean;

const
  HTTP_QUERY_CONTENT_LENGTH                   = 5;


  INTERNET_OPEN_TYPE_PRECONFIG        = 0;  { use registry configuration }
  INTERNET_FLAG_PRAGMA_NOCACHE                = $00000100; { asking wininet to add "pragma: no-cache" }
  INTERNET_FLAG_IGNORE_CERT_CN_INVALID        = $00001000; { bad common name in X509 Cert. }
  INTERNET_FLAG_IGNORE_CERT_DATE_INVALID      = $00002000; { expired X509 Cert. }
  INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS      = $00004000; { ex: http:// to https:// }
  INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP       = $00008000; { ex: https:// to http:// }
  INTERNET_FLAG_KEEP_CONNECTION               = $00400000; { use keep-alive semantics }
  INTERNET_FLAG_SECURE                        = $00800000; { use PCT/SSL if applicable (HTTP) }
  INTERNET_FLAG_RELOAD                        = $80000000; { retrieve the original item }

  INTERNET_ERROR_BASE                         = 12000;
  ERROR_INTERNET_INCORRECT_PASSWORD           = INTERNET_ERROR_BASE + 14;
  ERROR_INTERNET_FORCE_RETRY                  = INTERNET_ERROR_BASE + 32;
  ERROR_INTERNET_SEC_CERT_DATE_INVALID        = INTERNET_ERROR_BASE + 37;
  ERROR_INTERNET_SEC_CERT_CN_INVALID          = INTERNET_ERROR_BASE + 38;
  ERROR_INTERNET_HTTP_TO_HTTPS_ON_REDIR       = INTERNET_ERROR_BASE + 39;
  ERROR_INTERNET_HTTPS_TO_HTTP_ON_REDIR       = INTERNET_ERROR_BASE + 40;
  ERROR_INTERNET_POST_IS_NON_SECURE           = INTERNET_ERROR_BASE + 43;
  ERROR_INTERNET_INVALID_CA                   = INTERNET_ERROR_BASE + 45;

  FLAGS_ERROR_UI_FILTER_FOR_ERRORS            = $01;
  FLAGS_ERROR_UI_FLAGS_CHANGE_OPTIONS         = $02;
  FLAGS_ERROR_UI_FLAGS_GENERATE_DATA          = $04;

  INTERNET_INVALID_PORT_NUMBER = 0;                 { use the protocol-specific default }
  INTERNET_DEFAULT_HTTP_PORT = 80;                  {    "     "  HTTP   " }
  INTERNET_DEFAULT_HTTPS_PORT = 443;                {    "     "  HTTPS  " }

  INTERNET_SERVICE_HTTP = 3;

{ Flags for InternetGetConnectedState }
  INTERNET_CONNECTION_MODEM           = 1;
  INTERNET_CONNECTION_LAN             = 2;
  INTERNET_CONNECTION_PROXY           = 4;
  INTERNET_CONNECTION_MODEM_BUSY      = 8;

type
  HINTERNET = Pointer;
  INTERNET_PORT = Word; 

type
  TInternetOpenW = function (lpszAgent: PWideChar; dwAccessType: DWORD;
    lpszProxy, lpszProxyBypass: PWideChar; dwFlags: DWORD): HINTERNET; stdcall;
  TInternetOpenUrlW = function (hInet: HINTERNET; lpszUrl: PWideChar;
    lpszHeaders: PWideChar; dwHeadersLength: DWORD; dwFlags: DWORD;
      dwContext: DWORD): HINTERNET; stdcall;
  THttpOpenRequestW = function (hConnect: HINTERNET; lpszVerb: PWideChar;
    lpszObjectName: PWideChar; lpszVersion: PWideChar; lpszReferrer: PWideChar;
      lplpszAcceptTypes: PLPWSTR; dwFlags: DWORD;
        dwContext: DWORD): HINTERNET; stdcall;
  THttpSendRequestW = function (hRequest: HINTERNET; lpszHeaders: PWideChar;
    dwHeadersLength: DWORD; lpOptional: Pointer;
      dwOptionalLength: DWORD): BOOL; stdcall;
  THttpQueryInfoW = function(hRequest: HINTERNET; dwInfoLevel: DWORD;
    lpvBuffer: Pointer; var lpdwBufferLength: DWORD;
      var lpdwReserved: DWORD): BOOL; stdcall;
  TInternetConnectW = function (hInet: HINTERNET; lpszServerName: PWideChar;
    nServerPort: INTERNET_PORT; lpszUsername: PWideChar; lpszPassword: PWideChar;
      dwService: DWORD; dwFlags: DWORD; dwContext: DWORD): HINTERNET; stdcall;

  TInternetOpenA = function (lpszAgent: PAnsiChar; dwAccessType: DWORD;
    lpszProxy, lpszProxyBypass: PAnsiChar; dwFlags: DWORD): HINTERNET; stdcall;
  TInternetOpenUrlA = function (hInet: HINTERNET; lpszUrl: PAnsiChar;
    lpszHeaders: PAnsiChar; dwHeadersLength: DWORD; dwFlags: DWORD;
      dwContext: DWORD): HINTERNET; stdcall;
  THttpOpenRequestA = function (hConnect: HINTERNET; lpszVerb: PAnsiChar;
    lpszObjectName: PAnsiChar; lpszVersion: PAnsiChar; lpszReferrer: PAnsiChar;
      lplpszAcceptTypes: PLPSTR{TNT-ALLOW PLPSTR}; dwFlags: DWORD;
        dwContext: DWORD): HINTERNET; stdcall;
  THttpSendRequestA = function (hRequest: HINTERNET; lpszHeaders: PAnsiChar;
    dwHeadersLength: DWORD; lpOptional: Pointer;
      dwOptionalLength: DWORD): BOOL; stdcall;
  THttpQueryInfoA = function(hRequest: HINTERNET; dwInfoLevel: DWORD;
    lpvBuffer: Pointer; var lpdwBufferLength: DWORD;
      var lpdwReserved: DWORD): BOOL; stdcall;
  TInternetConnectA = function (hInet: HINTERNET; lpszServerName: PAnsiChar;
    nServerPort: INTERNET_PORT; lpszUsername: PAnsiChar; lpszPassword: PAnsiChar;
      dwService: DWORD; dwFlags: DWORD; dwContext: DWORD): HINTERNET; stdcall;

  TInternetQueryDataAvailable = function (hFile: HINTERNET; var lpdwNumberOfBytesAvailable: DWORD;
    dwFlags, dwContext: DWORD): BOOL; stdcall;
  TInternetReadFile = function (hFile: HINTERNET; lpBuffer: Pointer;
    dwNumberOfBytesToRead: DWORD; var lpdwNumberOfBytesRead: DWORD): BOOL; stdcall;
  TInternetCloseHandle = function (hInet: HINTERNET): BOOL; stdcall;
  TInternetErrorDlg = function (hWnd: HWND; hRequest: HINTERNET;
    dwError, dwFlags: DWORD; var lppvData: Pointer): DWORD; stdcall;
  TInternetGetConnectedState = function(lpdwFlags: LPDWORD; dwReserved: DWORD): BOOL; stdcall;

var
  InternetOpenW:    TInternetOpenW    = nil;
  InternetOpenUrlW: TInternetOpenUrlW = nil;
  HttpOpenRequestW: THttpOpenRequestW = nil;
  HttpSendRequestW: THttpSendRequestW = nil;
  HttpQueryInfoW:   THttpQueryInfoW   = nil;
  InternetConnectW: TInternetConnectW = nil;

  InternetOpenA:    TInternetOpenA    = nil;
  InternetOpenUrlA: TInternetOpenUrlA = nil;
  HttpOpenRequestA: THttpOpenRequestA = nil;
  HttpSendRequestA: THttpSendRequestA = nil;
  HttpQueryInfoA:   THttpQueryInfoA   = nil;
  InternetConnectA: TInternetConnectA = nil;

  InternetQueryDataAvailable: TInternetQueryDataAvailable = nil;
  InternetReadFile: TInternetReadFile = nil;
  InternetCloseHandle: TInternetCloseHandle = nil;
  InternetErrorDlg: TInternetErrorDlg = nil;
  InternetGetConnectedState: TInternetGetConnectedState = nil;

resourcestring
  SDefaultInternetConnectionDisabled = 'Connection to Internet is currently disabled.';

var
  WinInetIsLoaded: Boolean = False;
  WinInetIsUnicode: Boolean = False;
  SInternetConnectionDisabled: AnsiString = SDefaultInternetConnectionDisabled;

function UseWinInet(RootKey: HKEY; const RegKey, RegValue: WideString): Boolean;
function GetRegUseInternet(RootKey: HKEY; const RegKey, RegValue: WideString): Boolean;
procedure SetRegUseInternet(RootKey: HKEY; const RegKey, RegValue: WideString; Value: Boolean);
function GetHttpQueryInfoString(h_Request: HINTERNET; dwInfoLevel: DWORD): WideString;

procedure OpenURL(Url: WideString);
procedure OpenUserURL(Url: WideString);

implementation

uses
  Controls, Dialogs, TntDialogs, SysUtils, syncobjs,
  TntSysUtils, TntWindows,
  TntLXUtils, TntLXRegistry, TntLXVerUtils, TntLXRichEdits;

var
  WinInetDll: HModule = 0;

resourcestring
  SWinInetLoadError = 'Error trying to load WinInet.';
  SError = 'Error';

function GetRegUseInternet(RootKey: HKEY; const RegKey, RegValue: WideString): Boolean;
var
  Reg: TTntRegistryLX;
begin
  { Assume TRUE.  Registry must prove FALSE. }
  Result := True;
  if RegKey <> '' then begin
    Reg := TTntRegistryLX.Create;
    try
      Reg.Access := KEY_READ;
      Reg.RootKey := RootKey;
      if Reg.OpenKey(RegKey, False) then
        Result := Reg.ReadBoolDef(RegValue, Result);
    finally
      Reg.Free;
    end;
  end;
end;

procedure SetRegUseInternet(RootKey: HKEY; const RegKey, RegValue: WideString; Value: Boolean);
var
  Reg: TTntRegistryLX;
begin
  Reg := TTntRegistryLX.Create;
  try
    Reg.RootKey := RootKey;
    Reg.GotoKey(RegKey, True);
    Reg.WriteBool(RegValue, Value);
  finally
    Reg.Free;
  end;
end;

resourcestring
  SWhereToDownload = 'You can download it from %s';
  SIE5orNewer = 'Internet Explorer 5.0 or newer';
  SIEWebLink = 'http://www.microsoft.com/ie';

function UseWinInet(RootKey: HKEY; const RegKey, RegValue: WideString): Boolean;
begin
  result := GetRegUseInternet(RootKey, RegKey, RegValue);
  if (not result) then begin
    WinInetIsLoaded := False;
    InternetOpenW := nil;
    InternetOpenUrlW := nil;
    HttpOpenRequestW := nil;
    HttpSendRequestW := nil;
    HttpQueryInfoW := nil;
    InternetConnectW := nil;

    InternetOpenA := nil;
    InternetOpenUrlA := nil;
    HttpOpenRequestA := nil;
    HttpSendRequestA := nil;
    HttpQueryInfoa := nil;
    InternetConnectA := nil;

    InternetQueryDataAvailable := nil;
    InternetReadFile := nil;
    InternetCloseHandle := nil;
    InternetErrorDlg := nil;
    InternetGetConnectedState := nil;
  end else begin
    try
      if WinInetDll = 0 then
        WinInetDll := WinCheckH(Tnt_LoadLibraryW('wininet'));
      InternetOpenW := WinCheckP(GetProcAddress(WinInetDll, 'InternetOpenW'));
      InternetOpenUrlW := WinCheckP(GetProcAddress(WinInetDll, 'InternetOpenUrlW'));
      HttpOpenRequestW := WinCheckP(GetProcAddress(WinInetDll, 'HttpOpenRequestW'));
      HttpSendRequestW := WinCheckP(GetProcAddress(WinInetDll, 'HttpSendRequestW'));
      HttpQueryInfoW := WinCheckP(GetProcAddress(WinInetDll, 'HttpQueryInfoW'));
      InternetConnectW := WinCheckP(GetProcAddress(WinInetDll, 'InternetConnectW'));

      InternetOpenA := WinCheckP(GetProcAddress(WinInetDll, 'InternetOpenA'));
      InternetOpenUrlA := WinCheckP(GetProcAddress(WinInetDll, 'InternetOpenUrlA'));
      HttpOpenRequestA := WinCheckP(GetProcAddress(WinInetDll, 'HttpOpenRequestA'));
      HttpSendRequestA := WinCheckP(GetProcAddress(WinInetDll, 'HttpSendRequestA'));
      HttpQueryInfoa := WinCheckP(GetProcAddress(WinInetDll, 'HttpQueryInfoA'));
      InternetConnectA := WinCheckP(GetProcAddress(WinInetDll, 'InternetConnectA'));

      InternetQueryDataAvailable := WinCheckP(GetProcAddress(WinInetDll, 'InternetQueryDataAvailable'));
      InternetReadFile := WinCheckP(GetProcAddress(WinInetDll, 'InternetReadFile'));
      InternetCloseHandle := WinCheckP(GetProcAddress(WinInetDll, 'InternetCloseHandle'));
      InternetErrorDlg := WinCheckP(GetProcAddress(WinInetDll, 'InternetErrorDlg'));
      InternetGetConnectedState := WinCheckP(GetProcAddress(WinInetDll, 'InternetGetConnectedState'));

      WinInetIsLoaded := True;
      WinInetIsUnicode := GetMajorVersion(WinInetVersion) >= 5;
    except
      on E: Exception do begin
        MessageBoxW(0,
           PWideChar( WideFormat(SWinInetLoadError
                  + CRLF + SProgramRequires
                  + CRLF + SWhereToDownload, [SIE5orNewer, SIEWebLink])
          + CRLF
          + CRLF + E.Message), PWideChar(WideString(SError)), 0);
        raise;
      end;
    end;
  end;
end;

function InternetIsConnected: Boolean;
var
  Flags: DWord;
begin
  Flags := INTERNET_CONNECTION_MODEM or INTERNET_CONNECTION_LAN or INTERNET_CONNECTION_PROXY;
  result := WinInetIsLoaded and InternetGetConnectedState(@Flags, 0);
end;

function GetHttpQueryInfoString(h_Request: HINTERNET; dwInfoLevel: DWORD): WideString;
var
  NumChars: Cardinal;
  Dummy: Cardinal;
  Result_Ansi: AnsiString;
begin
  Dummy := 0;
  NumChars := 0;
  if WinInetIsUnicode then begin
    HttpQueryInfoW(h_Request, dwInfoLevel, nil, NumChars, Dummy);
    SetLength(Result, NumChars);
    Dummy := 0;
    HttpQueryInfoW(h_Request, dwInfoLevel, PWideChar(Result), NumChars, Dummy);
  end else begin
    HttpQueryInfoA(h_Request, dwInfoLevel, nil, NumChars, Dummy);
    SetLength(Result_Ansi, NumChars);
    Dummy := 0;
    HttpQueryInfoA(h_Request, dwInfoLevel, PAnsiChar(Result_Ansi), NumChars, Dummy);
    Result := Result_Ansi;
  end;
end;

procedure OpenURL(Url: WideString);
begin
  if WinInetIsLoaded then
    ShellOpenFile(Url)
  else
    raise ETntGeneralError.Create(SInternetConnectionDisabled);
end;

procedure OpenUserURL(Url: WideString);
begin
  if (not WinInetIsLoaded)
  and (WideTextPos('http', Url) = 1)
  and (WideMessageDlg('This action will cause your Web browser to connect to the Internet.'
    + CRLF + 'Do you want to proceed?', mtConfirmation, mbYesNoCancel, -1) <> mrYes)
  then
    Abort;
  ShellOpenFile(Url)
end;

initialization

finalization
  if WinInetDll <> 0 then
    FreeLibrary(WinInetDll);

end.
